import numpy as np

from ..match_detections import MatchEngineIoU


def add_confidence_from_max_iou(detections, ground_truths):
    r"""Compute confidence scores for detections based on the maximum IoU between detections and ground truths.

    Args:
        detections (ndarray, list) : A ndarray of detections stored as:

            * Bounding boxes for a given class where each row is a detection stored as:
              ``[BoundingBox]``
            * Polygons for a given class where each row is a detection stored as:
              ``[Polygon]``

        ground_truths (ndarray,list) : A ndarray of ground truth stored as:

            * Bounding boxes for a given class where each row is a ground truth stored as:
              ``[BoundingBox]``
            * Polygons for a given class where each row is a ground truth stored as:
              ``[Polygon]``

    Returns:
        ndarray: The detection with an additional column containing maximum IoU between detections and ground truths
        to use as confidence scores, as:

        * Bounding boxes for a given class where each row is a detection stored as:
          ``[BoundingBox, confidence]``
        * Polygons for a given class where each row is a detection stored as:
          ``[Polygon, confidence]``


    Examples:
        With a bounding box input:

        >>> from playground_metrics.utils.geometry_utils import BoundingBox
        >>> detections = np.array([[BoundingBox(0, 0, 9, 5)],
        ...                        [BoundingBox(23, 13, 29, 18)]])
        >>> ground_truths = np.array([[BoundingBox(5, 2, 15, 9)],
        ...                           [BoundingBox(18, 10, 26, 15)]])
        >>> add_confidence_from_max_IoU(detections, ground_truths)
        array([[BoundingBox(xmin=0, ymin=0, xmax=9, ymax=5), 0.11650485436893204],
               [BoundingBox(xmin=23, ymin=13, xmax=29, ymax=18), 0.09375]],
              dtype=object)

        The same with polygons:

        >>> from playground_metrics.utils.geometry_utils import Polygon
        >>> detections = np.array([[Polygon([[0.0, 0.0], [0.0, 5.0], [9.0, 5.0], [9.0, 0.0]])],
        ...                       [Polygon([[23.0, 13.0], [23.0, 18.0], [29.0, 18.0], [29.0, 13.0]])]])
        >>> ground_truths = np.array([[Polygon([[5, 2], [5, 9], [15, 9], [15, 2]])],
        ...                           [Polygon([[18, 10], [18, 15], [26, 15], [26, 10]])]])
        >>> add_confidence_from_max_IoU(detections, ground_truths)
        array([[Polygon(shell=[[0. 0.] [0. 5.] [9. 5.] [9. 0.]], holes=[]),
                0.11650485436893204],
               [Polygon(shell=[[23. 13.] [23. 18.] [29. 18.] [29. 13.]], holes=[]),
                0.09375]], dtype=object)

    """
    match_engine = MatchEngineIoU(0.5, 'coco')
    iou = match_engine.compute_similarity_matrix(np.insert(detections, 1, np.linspace(1, 0,
                                                                                      detections.shape[0]), axis=1),
                                                 ground_truths)
    return np.insert(detections, 1, np.max(iou, axis=1), axis=1)



from abc import ABC, abstractmethod
from collections.abc import MutableMapping

import numpy as np
import shapely.geometry

from ..exception import InvalidGeometryError


class AbstractBaseGeometry(ABC):
    r"""Abstract base class for every Geometry types.

    It gives exepcted valid interfaces of every geometric types,
    a pythonic repr and a convenient access to build-time parameters saved through ``**kwargs``.

    Subclasses must implement every abstract interfaces to be considered a valid class.

    Args:
        **kwargs (dict): Assumed as the build time coordinates arguments, passed to private attribute instance of
            :class:`~playground_metrics.utils.geometry_utils.base_geometry.GeometryCoordinates` to allow easy repr and
            attribute access (c.f. example).

    """

    __slots__ = '_coordinates'

    def __init__(self, **kwargs):
        super(AbstractBaseGeometry, self).__init__()
        self._coordinates = GeometryCoordinates(**kwargs)

    def __repr__(self):
        """Represent the geometry and its coordinates."""
        return '{}{}({})'.format('Empty' if self.is_empty else '', self.__class__.__name__, self._coordinates)

    def __dir__(self):
        """Return dir of geometry and its coordinates."""
        return list(super(AbstractBaseGeometry, self).__dir__()) + list(self._coordinates.keys())

    def __getattr__(self, item):
        r"""Allow coordinate access from outside as well as shapely internal attributes access.

        Not that overriding __getattr__ instead of __getattribute__ allows actual attributes to take precedence over
        dynamic attributes mapping.
        """
        try:
            return self._coordinates.__getattr__(item)
        except AttributeError:
            try:
                return self._internal.__getattribute__(item)
            except AttributeError:
                raise AttributeError('\'{}\' object has no attribute \'{}\''.format(self.__class__.__name__, item))

    @property
    @abstractmethod
    def is_empty(self):
        r"""bool: Whether the feature's `interior` and `boundary` (in point set terms) coincide with the empty set."""
        return NotImplemented

    @property
    @abstractmethod
    def area(self):
        r"""float: The area of the object."""
        return NotImplemented

    @property
    @abstractmethod
    def bounds(self):
        r"""(float, float, float, float): A ``(minx, miny, maxx, maxy)`` tuple of coordinates that bound the object."""
        return NotImplemented

    @property
    @abstractmethod
    def centroid(self):  # noqa: D205,D400
        r""":class:`~playground_metrics.utils.geometry_utils.geometry.Point`:
        The objects centroid as a Geometry type object.
        """
        return NotImplemented

    def __eq__(self, other):  # noqa: D205,D400
        r"""Return ``True`` if the two objects are of the same geometric type, and the coordinates of the two objects
        match precisely.

        Args:
            other (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometry to test equality with

        Returns:
            bool: Whether the 2 geometries are equal

        """
        if isinstance(other, AbstractBaseGeometry):
            return self.__class__ == other.__class__ and self._coordinates.__eq__(other._coordinates)
        else:
            return NotImplemented

    @abstractmethod
    def equals(self, other):  # noqa: D205,D400
        r"""Return ``True`` if the set-theoretic `boundary`, `interior`, and `exterior` of the object coincide with
        those of the other.

        Args:
            other (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometry to test equality with

        Returns:
            bool: Whether the subsets of the 2D plane described by the 2 geometries are equal

        """
        return NotImplemented

    @abstractmethod
    def distance(self, *others):  # noqa: D205,D400
        r"""Return the euclidean distance between the object centroid and the `other` object centroid or
        a tuple of distance if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute centroid distance with

        Returns:
            (float, tuple): The centroid distance or a tuple of centroid distance depending on the number of other
            geometries

        """
        return NotImplemented

    @abstractmethod
    def intersection(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection of this object with the `other` geometric object or a tuple
        of intersection if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection with

        Returns:
            (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`, tuple): The intersection
            or a tuple of intersection depending on the number of other geometries

        """
        return NotImplemented

    @abstractmethod
    def intersection_over_union(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection-over-union of this object with the `other` geometric object or
        a tuple of intersection-over-union if multiple input were given (c.f. :ref:`iou`).

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection over union with

        Returns:
            (float, tuple): The intersection over union or a tuple of intersection over union depending on the number
            of other geometries

        """
        return NotImplemented


class BaseGeometry(AbstractBaseGeometry):
    r"""Abstract base class for every Geometry types.

    It is merely a wrapper around :mod:`shapely.geometry` which
    implements a few differences in some binary operation, an intersection-over-union operator, a pythonic repr and
    a convenient access to build-time parameters saved through ``**kwargs``.

    Subclasses must implement the
    :meth:`~playground_metrics.utils.geometry_utils.base_geometry.BaseGeometry.from_shapely` class method to be
    instantiable.

    Args:
        shapely_internal (:class:`shapely.geometry.BaseGeometry`): A shapely
            :class:`~shapely.geometry.BaseGeometry` which will serve as the backend object for every geometric
            operations performed later on.
        **kwargs (dict): Assumed as the build time coordinates arguments, passed to private attribute instance of
            :class:`~playground_metrics.utils.geometry_utils.base_geometry.GeometryCoordinates` to allow easy repr and
            attribute access (c.f. example).

    Raises:
        ValueError: If the provided shapely geometry is not shapely.geometry geometry.
        :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`: If the provided geometry is deemed not
            valid by shapely.

    Examples:
        >>> class Point(BaseGeometry):
        ...    def __init__(self, x, y):
        ...        shapely_internal = shapely.geometry.Point(x, y)
        ...        super(Point, self).__init__(shapely_internal, x=x, y=y)
        ...
        ...    @classmethod   # To allow instantiation
        ...    def from_shapely(cls, geom):  # This is not a valid implementation
        ...        pass
        >>> point = Point(4, 6)
        >>> point
        Point(x=4, y=6)
        >>> point.x
        4
        >>> point.y
        6
        >>> print(point._internal)  # The underlying shapely object passed as shapely_internal to the constructor
        POINT (4 6)

    """

    __slots__ = '_internal',

    def __init__(self, shapely_internal, **kwargs):
        super(BaseGeometry, self).__init__()
        if not isinstance(shapely_internal, shapely.geometry.base.BaseGeometry):
            raise ValueError('BaseGeometry internal is not a valid '
                             'shapely.BaseGeometry instance '
                             '(found {} instead)'.format(shapely_internal.__class__.__name__))
        self._coordinates = GeometryCoordinates(**kwargs)
        if not shapely_internal.is_valid and not shapely_internal.bounds == (np.inf, np.inf, np.inf, np.inf):
            raise InvalidGeometryError('Invalid shapely.BaseGeometry created from '
                                       'provided parameters ({})'.format(self._coordinates))
        self._internal = shapely_internal

    @property
    def is_empty(self):
        r"""bool: Whether the feature's `interior` and `boundary` (in point set terms) coincide with the empty set."""
        return self._internal.is_empty

    @property
    def area(self):
        r"""float: The area of the object."""
        return self._internal.area

    @property
    def bounds(self):
        r"""(float, float, float, float): A ``(minx, miny, maxx, maxy)`` tuple of coordinates that bound the object."""
        return self._internal.bounds

    @property
    def centroid(self):  # noqa: D205,D400
        r""":class:`~playground_metrics.utils.geometry_utils.geometry.Point`:
        The objects centroid as a Geometry type object
        """
        return self.from_shapely(self._internal.centroid)

    def equals(self, other):  # noqa: D205,D400
        r"""Return ``True`` if the set-theoretic `boundary`, `interior`, and `exterior` of the object coincide with
        those of the other.

        Args:
            other (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometry to test equality with

        Returns:
            bool: Whether the subsets of the 2D plane described by the 2 geometries are equal.

        """
        return self._internal.equals(other)

    def distance(self, *others):  # noqa: D205,D400
        r"""Return the euclidean distance between the object centroid and the `other` object centroid or
        a tuple of distance if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute centroid distance with

        Returns:
            (float, tuple): The centroid distance or a tuple of centroid distance depending on the number of other
            geometries

        """
        if len(others) == 1:
            return self._internal.centroid.distance(others[0].centroid)
        else:
            return tuple(self._internal.centroid.distance(other.centroid) for other in others)

    def _intersection(self, other):
        return self._internal.intersection(other)

    def intersection(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection of this object with the `other` geometric object or a tuple
        of intersection if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection with

        Returns:
            (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`, tuple): The intersection
            or a tuple of intersection depending on the number of other geometries

        """
        if len(others) == 1:
            return self.from_shapely(self._intersection(others[0]))
        else:
            return tuple(self.from_shapely(self._intersection(other)) for other in others)

    def intersection_over_union(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection-over-union of this object with the `other` geometric object or
        a tuple of intersection-over-union if multiple input were given (c.f. :ref:`iou`).

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection over union with

        Returns:
            (float, tuple): The intersection over union or a tuple of intersection over union depending on the number
            of other geometries

        """
        out = [None] * len(others)
        if len(others) == 1:
            inters = (self._intersection(others[0]), )
        else:
            inters = tuple(self._intersection(other) for other in others)

        for i, (other, inter) in enumerate(zip(others, inters)):

            if inter.is_empty or inter.area == 0:
                out[i] = 0.0
                continue

            # Compute the intersection over union by taking the intersection area and dividing it by the sum of
            # prediction + ground-truth areas - the intersection area
            iou = inter.area / float(self.area + other.area - inter.area)

            # Return the intersection over union value
            out[i] = iou

        if len(out) == 1:
            return out[0]
        else:
            return tuple(out)

    @classmethod
    @abstractmethod
    def from_shapely(cls, geom):
        r"""Perform the inverse mapping from a shapely object to a Geometry type.

        To be more precise, :class:`~playground_metrics.utils.geometry_utils.base_geometry.BaseGeometry` subclasses
        implement a forward mapping from a Geometry type to a shapely object. However, because geometric operations are
        not guaranteed to keep the Geometric type (e.g. intersection between a Polygon and a Point is a Point, not a
        Polygon), and because all the actual operations are performed by shapely, it is necessary to be able to perform
        the backward mapping to transform an operation result back into valid Geometry type.

        The method was made left to subclasses to implement to avoid the base class to be explicitly aware of its
        subclasses, which would be bad design.

        Args:
            geom (:class:`shapely.geometry.BaseGeometry`): The shapely :class:`~shapely.geometry.BaseGeometry` to be
                converted into a :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` object.

        Returns:
            :class:`~map_metric_api.utils.geometry_utils.base_geometry.BaseGeometry`: The backward mapping resulting
            Geometry type object.

        """
        raise NotImplementedError


class GeometryCoordinates(MutableMapping):
    r"""A container for coordinates with dict-like interface.

    It stores every keyword arguments passed to its constructor and makes them accessible as attributes as well.

    It is used internally by :class:`~playground_metrics.utils.geometry_utils.base_geometry.BaseGeometry` to delegate
    some coordinates specific function or representation.

    """

    __slots__ = '_internal',

    def __init__(self, **kwargs):
        super(GeometryCoordinates, self).__setattr__('_internal', kwargs)

    def __eq__(self, other):
        """Return True if two GeometryCoordinates instances have the same coordinates."""
        if isinstance(other, GeometryCoordinates):
            if list(self.keys()) == list(other.keys()):
                for k in self.keys():
                    if isinstance(self[k], np.ndarray) and isinstance(other[k], np.ndarray):
                        if not np.all(self[k] == other[k]):
                            return False
                    elif not self[k] == other[k]:
                        return False
                return True
            else:
                return False
        else:
            return NotImplemented

    def __repr__(self):
        """Return the coordinate as a comma-separated key=value format string."""
        return ' '.join(', '.join(('{}={}'.format(key, value)
                                   for key, value in self._internal.items())).replace('\n', '').split())

    def __dir__(self):
        """Return the coordinate dir."""
        return super(GeometryCoordinates, self).__dir__() + list(self._internal.keys())

    def __getattr__(self, key):
        """Get a coordinate value."""
        try:
            return self._internal[key]
        except KeyError:
            raise AttributeError('\'GeometryCoordinates\' object has no attribute \'{}\''.format(key))

    def __setattr__(self, key, value):
        """Add or set a coordinate."""
        self._internal[key] = value

    def __delattr__(self, key):
        """Delete a coordinate."""
        del self._internal[key]

    def __setitem__(self, key, value):
        """Add or set a coordinate."""
        self._internal[key] = value

    def __delitem__(self, key):
        """Delete a coordinate."""
        del self._internal[key]

    def __getitem__(self, key):
        """Get a coordinate value."""
        return self._internal[key]

    def __len__(self):
        """Return length of coordinates array."""
        return len(self._internal)

    def __iter__(self):
        """Iterate over coordinates."""
        return iter(self._internal)

    def keys(self):
        r"""Return a generator over coordinates name."""
        return self._internal.keys()



"""Implement the public interface to compute mAP from a set of detections and ground truths.

For more in formation on how the metric is computed, see: :doc:`../content/map_metric`.

If one wants to integrate the module into a framework to use it as a validation metric, the
:class:`~MeanAveragePrecisionMetric` class described below should be wrappped accordingly to follow the framework
convention.
"""
import warnings
from collections import defaultdict

import numpy as np

from .utils import to_builtin, to_list
from .match_detections import MatchEngineIoU
from .utils.geometry_utils import get_type_and_convert


class MeanAveragePrecisionMetric:
    r"""Implement an API to compute mAP.

    It gives three methods:

        * :meth:`update(detections, ground_truths) <update>` which accumulates TP, FP and FN over examples
        * :meth:`compute` which computes mAP and AP per label from accumulated values
        * :meth:`reset` which resets accumulated values to their initial values to start mAP computation from scratch

    See Also:
        Information on how **mAP**, **AP**, **precision** and **recall** are computed may be found in
        :doc:`../content/map_metric`.

    Args:
        threshold (float): Optional, default to 0.5. Similarity threshold for which we consider a valid
            match between detection and ground truth.
        match_algorithm (str): Optional, default to 'coco'. 'xview' or 'coco' to choose the matching algorithm (c.f.
            :ref:`match`) or 'non-unitary' to use non-unitary matching.
        label_mean_area (dict) : Optional, default to ``None``. A dictionary containing the mean area for each label in
            the dataset, if given, it is used to match with *iIoU* instead of *IoU* (c.f. :ref:`iiou`).
        trim_invalid_geometry (bool): Optional, default to ``False``. If set to ``True`` conversion will ignore invalid
            geometries and leave them out of mAP computations. This means that the function will work on arrays where
            ``work_array.shape[0] <= input_array.shape[0]``.  If set to ``False``, an invalid geometry will raise an
            :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`.
        autocorrect_invalid_geometry (Bool): Optional, default to ``False``. Whether to attempt correcting a faulty
            geometry to form a valid one. If set to ``True`` and the autocorrect attempt is unsuccessful, it falls back
            to the behaviour defined in ``trim_invalid_geometry``.
        match_engine (:class:`~map_metric_api.match_detections.MatchEngineBase`): Optional, default to
            :class:`~playground_metrics.match_detections.MatchEngineIoU`. If provided matching will be done using the
            provided ``match_engine`` instead of the default one. Note that the ``threshold`` and ``match_algorithm``
            provided parameters will be overridden by those provided in the ``match_engine``.

    Warning:
        When using non-unitary matching, the AP per class and the mAP are ill-defined and must be taken with a grain
        of salt.

    Warns:
        UserWarning: If ``match_algorithm`` is 'non-unitary' to warn that mAP and AP per class values are
            ill-defined.
        RuntimeWarning: If a ``match_engine`` is provided and its ``threshold`` or ``match_algorithm``
            attribute differs from those provided as arguments to the constructor.

    Note:
        * Polygon auto-correction only corrects self-crossing exterior rings, in which case it creates one Polygon
          out of every simple ring which might be extracted from the original Polygon exterior.
        * Polygon auto-correction will systematically fail on Polygons with at least one inner ring.

    Attributes:
        mAP (float) : The mAP computed by :meth:`compute` from accumulated values
        average_precision_per_class (defaultdict) : The AP for each label as constructed by :meth:`compute` from
            accumulated values
        precision_per_class (defaultdict) : The precision for each label as constructed by :meth:`compute` from
            accumulated values
        recall_per_class (defaultdict) : The recall for each label as constructed by :meth:`compute` from accumulated
            values
        number_true_detection_per_class (defaultdict): The number of detection matched to a ground truth as
            constructed by :meth:`compute` from accumulated values
        number_false_detection_per_class (defaultdict): The number of detection not matched to a ground truth as
            constructed by :meth:`compute` from accumulated value
        number_found_ground_truth_per_class (defaultdict): The number of ground truth matched to a detection as
            constructed by :meth:`compute` from accumulated values
        number_missed_ground_truth_per_class (defaultdict): The number of ground truth not matched to a detection as
            constructed by :meth:`compute` from accumulated values
        match_engine (:class:`~map_metric_api.match_detections.MatchEngineBase`) : The match_engine object used to match
            detections and ground truths. If none where provided in the constructor call, it defaults to
            :class:`~playground_metrics.match_detections.MatchEngineIoU`.

    """

    def __init__(self, threshold=None, match_algorithm=None, label_mean_area=None, trim_invalid_geometry=False,
                 autocorrect_invalid_geometry=False, match_engine=None):

        if match_engine is not None and (threshold is not None or match_algorithm is not None):
            warnings.warn('In the future match_engine will be made incompatible with threshold and match_algorithm. '
                          'Providing both will raise a ValueError.', FutureWarning)

        # Set configurations values
        threshold = threshold if threshold is not None else 0.5
        match_algorithm = match_algorithm or 'coco'
        self.match_engine = match_engine or MatchEngineIoU(threshold, match_algorithm)
        if threshold != self.threshold:
            warnings.warn('Discrepancy between user provided threshold and '
                          'match_engine threshold ({} != {})'.format(threshold, self.threshold), RuntimeWarning)

        if match_algorithm != self.match_engine.match_algorithm:
            warnings.warn('Discrepancy between user provided match_algorithm and '
                          'match_engine match_algorithm ({} != {})'.format(match_algorithm,
                                                                           self.match_engine.match_algorithm),
                          RuntimeWarning)

        if match_algorithm == 'non-unitary':
            warnings.warn('When using non-unitary matching, the AP per class and the mAP are '
                          'ill-defined and must be taken with a grain of salt.', UserWarning)

        self.label_mean_area = label_mean_area
        self.trim_invalid_geometry = trim_invalid_geometry
        self.autocorrect_invalid_geometry = autocorrect_invalid_geometry
        # Set intermediate and return values
        self._init_values()

    @property
    def threshold(self):  # noqa: D205,D400
        """float: The IoU threshold by :attr:`self.match_engine <match_engine>` or ``None``
        if :attr:`self.match_engine <match_engine>` doesn't use any threshold.
        """
        try:
            return self.match_engine.threshold
        except AttributeError:
            return None

    @property
    def ground_truth_labels(self):
        """set: The set of unique label accumulated up to this point."""
        return self._ground_truth_labels

    def _init_values(self):
        # Set intermediate values
        self._detection_matched = defaultdict(self._empty_array)
        self._ground_truth_matched = defaultdict(self._empty_array)
        self._number_of_ground_truths = defaultdict(int)
        self._ground_truth_labels = set()
        self._confidence = defaultdict(self._empty_array)
        # Set return values
        self.mAP = 0.0  # pylint: disable=invalid-name
        self.average_precision_per_class = defaultdict(float)
        self.precision_per_class = defaultdict(float)
        self.recall_per_class = defaultdict(float)
        self.number_true_detection_per_class = defaultdict(int)
        self.number_false_detection_per_class = defaultdict(int)
        self.number_found_ground_truth_per_class = defaultdict(int)
        self.number_missed_ground_truth_per_class = defaultdict(int)

    def update(self, detections, ground_truths):
        r"""Accumulate values necessary to compute mAP with detections and ground truths of a single image.

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[x_min, y_min, x_max, y_max, confidence, label]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[[[outer_ring], [inner_rings]], confidence, label]``
                * Points for a given class where each row is a detection stored as:
                  ``[x, y, confidence, label]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[x_min, y_min, x_max, y_max, label]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[[[outer_ring], [inner_rings]], label]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[x, y, label]``

        Raises:
            KeyError : If ``self.label_mean_area`` is not ``None`` but a label is missing

        The input detections and ground truths are allowed to be **points** in the documentation.
        This is to allow the use of a custom `MatchEngine` for points, however, the default
        :class:`~playground_metrics.match_detections.MatchEngineIoU` works on **intersection-over-union** which
        is incompatible with **points**. More information on input geometrical types can be found in
        :doc:`playground_metrics.match_detections`.

        Note:
            The labels provided in the input arrays can theoretically be any hashable type, however,
            only numeric types, strings and tuples are officially supported.

        """
        detections_type, detections = self._format_input(detections)
        ground_truths_type, ground_truths = self._format_input(ground_truths)

        if detections.size == ground_truths.size == 0:
            return

        if detections.size == 0:
            self._ground_truth_labels.update(to_list(ground_truths[:, 1]))
            for ground_truth_label in self._ground_truth_labels:
                self._number_of_ground_truths[ground_truth_label] += \
                    len(ground_truths[to_builtin(ground_truths[:, 1]) == ground_truth_label, :1])
                self._ground_truth_matched[ground_truth_label] = \
                    np.concatenate((self._ground_truth_matched[ground_truth_label],
                                    np.zeros((ground_truths[to_builtin(ground_truths[:, 1]) == ground_truth_label,
                                                            :].shape[0]))))
            return

        if ground_truths.size == 0:
            self._ground_truth_labels.update(to_list(detections[:, 2]))
            for ground_truth_label in self._ground_truth_labels:
                self._detection_matched[ground_truth_label] = np.concatenate(
                    (self._detection_matched[ground_truth_label],
                     np.zeros((detections[to_builtin(detections[:, 2]) == ground_truth_label, :2].shape[0]))))
                self._confidence[ground_truth_label] = \
                    np.concatenate((self._confidence[ground_truth_label],
                                    np.sort(detections[to_builtin(detections[:, 2]) == ground_truth_label, 1])[::-1]))
            return

        self._ground_truth_labels.update(to_list(ground_truths[:, 1]), to_list(detections[:, 2]))

        for ground_truth_label in self._ground_truth_labels:
            try:
                mean_area = self.label_mean_area[ground_truth_label]
            except KeyError:
                raise KeyError('label_mean_area is missing the label {}'.format(ground_truth_label))
            except TypeError:
                mean_area = None

            match_matrix = \
                self.match_engine.match(detections[to_builtin(detections[:, 2]) == ground_truth_label, :2],
                                        ground_truths[to_builtin(ground_truths[:, 1]) == ground_truth_label, :1],
                                        label_mean_area=mean_area)

            # Having this before checking if there were detections for this particular class breaks the xview score
            # equality test, however this is the way to go to ensure that False-Negative are correctly accounted for
            # in all cases. So xView scoring code is wrong again here, sorry xView.
            self._number_of_ground_truths[ground_truth_label] += \
                len(ground_truths[to_builtin(ground_truths[:, 1]) == ground_truth_label, :1])

            self._ground_truth_matched[ground_truth_label] = \
                np.concatenate((self._ground_truth_matched[ground_truth_label], np.clip(match_matrix.sum(0), 0, 1)))

            # If no detections for this label pass here
            if match_matrix.shape[0] == 0:
                continue

            self._detection_matched[ground_truth_label] = \
                np.concatenate((self._detection_matched[ground_truth_label], np.clip(match_matrix.sum(1), 0, 1)))
            self._confidence[ground_truth_label] = \
                np.concatenate((self._confidence[ground_truth_label],
                                np.sort(detections[to_builtin(detections[:, 2]) == ground_truth_label, 1])[::-1]))

    def compute(self):
        r"""Compute the **mAP** according to the accumulated values.

        Moreover it sets the value for the following attributes:

            * :attr:`self.precision_per_class <precision_per_class>`: A dict of precisions per label
            * :attr:`self.recall_per_class <recall_per_class>`: A dict of recall per label
            * :attr:`self.average_precision_per_class <average_precision_per_class>`: A dict of average precisions
              per label
            * :attr:`self.number_true_detection_per_class <number_true_detection_per_class>`: A dict of the number
              of detection matched to a ground truth
            * :attr:`self.number_false_detection_per_class <number_false_detection_per_class>`: A dict of the number
              of detection not matched to a ground truth
            * :attr:`self.number_found_ground_truth_per_class <number_found_ground_truth_per_class>`: A dict of the
              number of ground truth matched to a detection
            * :attr:`self.number_missed_ground_truth_per_class <number_missed_ground_truth_per_class>`: A dict of the
              number of ground truth not matched to a detection

        Returns:
            float : The Mean Average Precision metric

        """
        for ground_truth_label in self._ground_truth_labels:

            # Compute the Det positive, Det negative, Gt positive and gt negative counters
            self.number_true_detection_per_class[ground_truth_label] = \
                np.sum(self._detection_matched[ground_truth_label]).item()
            self.number_false_detection_per_class[ground_truth_label] = \
                np.sum(np.logical_not(self._detection_matched[ground_truth_label])).item()
            self.number_found_ground_truth_per_class[ground_truth_label] = \
                np.sum(self._ground_truth_matched[ground_truth_label]).item()
            self.number_missed_ground_truth_per_class[ground_truth_label] = \
                np.sum(np.logical_not(self._ground_truth_matched[ground_truth_label])).item()

            if self._number_of_ground_truths[ground_truth_label] != 0:
                # Prepare the cumulative sum along confidence-sorted detections to compute Precision(Recall)
                sorted_detection_indices = np.argsort(self._confidence[ground_truth_label])[::-1]
                tp_sum = np.cumsum(self._detection_matched[ground_truth_label][sorted_detection_indices])
                fp_sum = np.cumsum(
                    np.logical_not(self._detection_matched[ground_truth_label][sorted_detection_indices])
                )

                # Compute the Precision(Recall) function
                precision = tp_sum / (tp_sum + fp_sum + np.spacing(1))
                recall = tp_sum / self._number_of_ground_truths[ground_truth_label]

                # Compute the precision and recall
                # For precision tp is the number of detections matched to the ground truth (unique, non-unique matches)
                tp = np.sum(self._detection_matched[ground_truth_label])
                p = len(self._detection_matched[ground_truth_label])
                self.precision_per_class[ground_truth_label] = (tp / p).item()
                # For recall tp is the number of ground-truth targets matched to detections
                tp = np.sum(self._ground_truth_matched[ground_truth_label])
                a = self._number_of_ground_truths[ground_truth_label]
                self.recall_per_class[ground_truth_label] = (tp / a).item()

                # Average precision and mAP computation
                precision, recall = self._remove_jaggedness(precision, recall)
                self.average_precision_per_class[ground_truth_label] = \
                    self._integrate_precision_recall_curve(precision, recall).item()
            else:
                self.precision_per_class[ground_truth_label] = 0.0
                self.recall_per_class[ground_truth_label] = np.nan

        # self.average_precision_per_class is a defaultdict, when label is absent the value is 0 by default
        self.mAP = np.nanmean(np.array([self.average_precision_per_class[label]
                                        for label in self._ground_truth_labels]))
        return self.mAP

    def reset(self):
        r"""Reset all intermediate and return values to their initial value.

        If :meth:`reset` is not called in-between two :meth:`compute` call, the values returned by :meth:`compute`
        will take into account the entire prediction stack, not just the predictions in-between the two
        :meth:`compute` calls.

        """
        self._init_values()

    def _format_input(self, input_array):
        return get_type_and_convert(input_array, trim_invalid_geometry=self.trim_invalid_geometry,
                                    autocorrect_invalid_geometry=self.autocorrect_invalid_geometry)

    @staticmethod
    def _remove_jaggedness(precision, recall):
        recall = np.concatenate([[0], recall, [1]])
        precision = np.concatenate([[0], precision, [0]])
        for i in range(precision.shape[0] - 2, 0, -1):
            if precision[i] > precision[i - 1]:
                precision[i - 1] = precision[i]
        return precision, recall

    @staticmethod
    def _integrate_precision_recall_curve(precision, recall):
        # The indices where recall changes value
        i = np.where(recall[1:] != recall[:len(recall) - 1])[0] + 1

        # Integration with step interpolation
        average_precision = np.sum((recall[i] - recall[i - 1]) * precision[i])

        return average_precision

    # Default factories
    @staticmethod
    def _nan():
        return np.nan

    @staticmethod
    def _empty_array():
        return np.array([])



def to_list(arr):
    """Transform an numpy array into a list with homogeneous element type."""
    return arr.astype(type(arr.ravel()[0])).tolist()


def to_builtin(arr):
    """Transform an numpy array into a numpy array with homogeneous element type."""
    return arr.astype(type(arr.ravel()[0]))



# Top-level exception
class MapMetricApiError(Exception):
    r"""Generic exception class for playground_metrics specific exceptions."""

    pass


# Match detections exceptions
class MatchEngineError(MapMetricApiError):
    r"""Generic exception class for MatchEngine specific exceptions."""

    pass


# Map metric exceptions
class MeanAveragePrecisionMetricError(MapMetricApiError):
    r"""Generic exception class for MeanAveragePrecisionMetric specific exceptions."""

    pass


# Geometry utils exceptions
class GeometryError(MapMetricApiError):
    r"""Generic exception class for geometry_utils specific exceptions."""

    pass


class ShapelySpecificTypeError(GeometryError):  # noqa: D205,D400
    r"""Raised whenever asked to handle a shapely.geometry type which have no equivalent type in geometry_utils
    (e.g. MultiPolygons)
    """

    pass


class InvalidGeometryError(GeometryError):
    r"""Raised when asked to create an invalid Geometry object."""

    pass


class InvalidGeometryOperationError(GeometryError):
    r"""Raised if an operation on Geometry returns an invalid Geometry."""

    pass



"""Implement the public interface to match a set of detections and ground truths."""

from abc import ABC, abstractmethod

import rtree
import numpy as np

from .utils.geometry_utils import BoundingBox, Polygon, Point


def _make_rtree(detections, transform_fn):
    """Make a fast RTree index from an iterable of detections.

    Args:
        detections (Iterable): An iterable of |Geometry| object.
        transform_fn (callable): A function taking a |Geometry| object an returning a box coordinates as a
            ``(minx, miny, maxx, maxy)`` tuple.

    Returns:
        rtree.index.Index: A RTree index populated with detections bounding box.

    """
    def enumerate_detections(iterable):
        for i, element in enumerate(iterable):
            yield i, transform_fn(element), None

    rtree_index_prop = rtree.index.Property()
    rtree_index_prop.fill_factor = 0.5
    rtree_index_prop.dimension = 2
    return rtree.index.Index(enumerate_detections(detections), properties=rtree_index_prop)


class MatchEngineBase(ABC):
    """Match detection with their ground truth according a similarity matrix and a detection confidence score.

    Matching may be done using coco algorithm or xView algorithm (which yield different matches as described for an
    intersection-over-union similarity matrix in :ref:`match`) or with non-unitary matching.

    Subclasses must implement :meth:`compute_similarity_matrix` and :meth:`trim_similarity_matrix` to be functional.

    Args:
        match_algorithm (str) : Either 'coco', 'xview' or 'non-unitary' to choose the match algorithm

    Attributes:
        match_algorithm (str) : Either 'coco', 'xview' or 'non-unitary' and indicates the match algorithm used

    """

    def __init__(self, match_algorithm):
        if match_algorithm not in ['coco', 'xview', 'non-unitary']:
            raise ValueError("match_algorithm must be either coco, xview or non-unitary")

        self.match_algorithm = match_algorithm

        # Authorized geometric types fot this match engine
        self._detection_types = (BoundingBox, Polygon, Point)
        self._ground_truth_types = (BoundingBox, Polygon, Point)

    def __repr__(self):
        """Represent the :class:`~playground_metrics.match_detections.MatchEngineBase` as a string."""
        d_arg = []
        for arg in ['threshold', 'match_algorithm', 'bounding_box_size']:
            if hasattr(self, arg):
                d_arg.append('{}={}'.format(arg, self.__getattribute__(arg)))
        return '{}({})'.format(self.__class__.__name__, ', '.join(d_arg))

    def __str__(self):
        """Represent the :class:`~playground_metrics.match_detections.MatchEngineBase` as a string."""
        d_arg = []
        for arg in ['threshold', 'match_algorithm', 'bounding_box_size']:
            if hasattr(self, arg):
                d_arg.append('{}={}'.format(arg, self.__getattribute__(arg)))
        return '{}({})'.format(self.__class__.__name__.replace('MatchEngine', ''), ', '.join(d_arg))

    def _compute_similarity_matrix_and_trim(self, detections, ground_truths, label_mean_area=None):
        similarity_matrix = self.compute_similarity_matrix(detections, ground_truths, label_mean_area)
        return similarity_matrix, self.trim_similarity_matrix(similarity_matrix, detections, ground_truths,
                                                              label_mean_area)

    @abstractmethod
    def compute_similarity_matrix(self, detections, ground_truths, label_mean_area=None):
        r"""Compute a similarity matrix between detections and ground truths.

        Abstract method.

        This method must be overridden in subsequent subclasses to handle both bounding box and polygon input format.

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray : A similarity matrix of dimension (#detections, #ground truth)

        """
        raise NotImplementedError

    @abstractmethod
    def trim_similarity_matrix(self, similarity_matrix, detections, ground_truths,
                               label_mean_area=None):  # noqa: D205,D400
        r"""Compute an array containing the indices in columns of similarity passing the first trimming (typically for
        IoU this would be the result of a simple thresholding) but it might be any method fit to do a rough filtering of
        possible ground truth candidates to match with a given detection.

        Abstract method.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray: An array of dimension (2, N) where each column is a tuple (detection, ground truth) describing
            a potential match. To be more precise, each match-tuple in the array corresponds to a position in the
            similarity matrix which will be used by the match algorithm to compute the final match.

        """
        raise NotImplementedError

    def match(self, detections, ground_truths, label_mean_area=None):  # noqa: D205,D400
        r"""Match detections :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` with ground truth
        :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` at a given similarity matrix and trim
        method using either Coco algorithm, xView algorithm or a naive *non-unitary* match.

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray : A binary matrix of all matches of dimension (#detections, #ground truth)

        """
        if detections.shape[0] == 0:
            return np.zeros((0, ground_truths.shape[0]))

        if ground_truths.shape[0] == 0:
            return np.zeros((detections.shape[0], 0))

        # Geometric static typing
        if not all((isinstance(geom, self._detection_types) for geom in detections[:, 0])):
            raise TypeError('Invalid geometric type provided in '
                            'detections, expected to be on of {}'
                            ''.format(' '.join(['{}'.format(geom_type.__name__)
                                                for geom_type in self._detection_types])))
        if not all((isinstance(geom, self._ground_truth_types) for geom in ground_truths[:, 0])):
            raise TypeError('Invalid geometric type provided in '
                            'detections, expected to be on of {}'
                            ''.format(' '.join(['{}'.format(geom_type.__name__)
                                                for geom_type in self._ground_truth_types])))

        # We sort detections by confidence before computing the similarity matrix
        detections = self._sort_detection_by_confidence(detections)

        # Compute similarity matrix and An array containing the indices in columns of similarity passing the first
        # trimming (Typically for IoU this would be the result of a simple thresholding).
        similarity_matrix, similarity_matches = self._compute_similarity_matrix_and_trim(detections, ground_truths,
                                                                                         label_mean_area)

        # We match the detection and the ground truth using the configured algorithm
        if self.match_algorithm == 'coco':
            return self._coco_match(similarity_matrix, similarity_matches)
        if self.match_algorithm == 'non-unitary':
            return self._non_unitary_match(similarity_matrix, similarity_matches)
        if self.match_algorithm == 'xview':
            return self._xview_match(similarity_matrix, similarity_matches)
        raise ValueError('Invalid match algorithm: must be either coco, xview or non-unitary')

    @staticmethod
    def _sort_detection_by_confidence(detections):
        # We sort the detection by decreasing confidence
        sort_indices = np.argsort(detections[:, 1])[::-1]
        return detections[sort_indices, :]

    @staticmethod
    def _coco_match(similarity_matrix, similarity_matches):  # noqa: D205,D400
        r"""Match detections bounding boxes with ground truth bounding boxes for a given similarity matrix and trim
        method using Coco algorithm.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)

        Returns:
            ndarray: A binary matrix of all matches of dimension (#detections, #ground truth)

        """
        # We prepare the detection match matrix
        match_matrix = np.zeros_like(similarity_matrix)

        if similarity_matches.shape[1] == 0:  # No matches at all
            return match_matrix

        forward = {match[0, 0]: match[1, :]
                   for match in np.hsplit(similarity_matches, np.where(np.diff(similarity_matches[0, :]) != 0)[0] + 1)}
        similarity_matches_by_gt = similarity_matches[:, np.argsort(similarity_matches[1, :])]
        backward = {match[1, 0]: match[0, :]
                    for match in np.hsplit(similarity_matches_by_gt,
                                           np.where(np.diff(similarity_matches_by_gt[1, :]) != 0)[0] + 1)}

        for k in range(similarity_matrix.shape[0]):
            # For each detection we select its ground truth match
            detection_matches = forward.get(k, np.zeros((0, 0)))

            # If we don't have anything left to match -> skip
            if detection_matches.size == 0:
                continue

            # We select the biggest similarity_matrix over them
            m = np.argmax(similarity_matrix[k, detection_matches])
            n = detection_matches[m]

            # We delete the ground truth column index from future match testing
            for d in backward[n]:
                forward[d] = forward[d][forward[d] != n]

            # We set the match flag to 1
            match_matrix[k, n] = 1

        return match_matrix

    @staticmethod
    def _xview_match(similarity_matrix, similarity_matches):  # noqa: D205,D400
        r"""Match detections bounding boxes with ground truth bounding boxes for a0 given similarity matrix and trim
        method using xView algorithm.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)

        Returns:
            ndarray: A binary matrix of all matches of dimension (#detections, #ground truth)

        """
        # We prepare the detection match matrix
        match_matrix = np.zeros_like(similarity_matrix)

        if similarity_matches.shape[1] == 0:  # No matches at all
            return match_matrix

        ground_truth_match_vector = [0] * similarity_matrix.shape[1]

        forward = {match[0, 0]: match[1, :]
                   for match in np.hsplit(similarity_matches, np.where(np.diff(similarity_matches[0, :]) != 0)[0] + 1)}

        for k in range(similarity_matrix.shape[0]):
            # For each detection we select its ground truth match
            detection_matches = forward.get(k, np.zeros((0, 0)))

            # If we don't have anything left to match -> skip
            if detection_matches.size == 0:
                continue

            # We select the biggest similarity_matrix over them
            m = np.argmax(similarity_matrix[k, detection_matches])
            n = detection_matches[m]

            if ground_truth_match_vector[n] == 0:
                # We match the detection and the ground truth
                ground_truth_match_vector[n] = 1
                match_matrix[k, n] = 1

        return match_matrix

    @staticmethod
    def _non_unitary_match(similarity_matrix, similarity_matches):  # noqa: D205,D400
        r"""Match detections bounding boxes with ground truth bounding boxes for a given similarity matrix for every
        positive example yielded by the  trim method.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)

        Returns:
            ndarray: A binary matrix of all matches of dimension (#detections, #ground truth)

        """
        # We prepare the detection match matrix
        match_matrix = np.zeros_like(similarity_matrix)
        match_matrix[similarity_matches[0, :], similarity_matches[1, :]] = 1

        return match_matrix


class MatchEngineIoU(MatchEngineBase):
    """Match detection with their ground truth according the their IoU and the detection confidence score.

    Args:
        threshold (float): The IoU threshold at which one considers a potential match as valid
        match_algorithm (str) : Either 'coco', 'xview' or 'non-unitary' to choose the match algorithm

    """

    def __init__(self, threshold, match_algorithm):
        super(MatchEngineIoU, self).__init__(match_algorithm)

        self._detection_types = (BoundingBox, Polygon)
        self._ground_truth_types = (BoundingBox, Polygon)

        self.threshold = threshold

    def _rtree_compute_iou_matrix(self, detections, ground_truths, label_mean_area=None):  # noqa: D205,D400
        r"""Compute the iou scores between all pairs of :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry`
        with an Rtree on detections to speed up computation.

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset,
                if given, it is used to match with *iIoU* instead of *IoU* (c.f. :ref:`iiou`)

        Returns:
            ndarray : An IoU matrix (#detections, #ground truth)

        """
        # We prepare the IoU matrix (#detection, #gt)
        iou = np.zeros((detections.shape[0], len(ground_truths)))

        detections = self._sort_detection_by_confidence(detections)

        # We construct a Rtree on detections
        def get_bounds(geometry):
            return geometry[0].bounds

        rtree_index = _make_rtree(detections, get_bounds)

        for k, ground_truth in enumerate(ground_truths):
            overlapping_detections = rtree_index.intersection(ground_truth[0].bounds, objects=False)
            for m in overlapping_detections:
                if label_mean_area is not None:
                    iou[m, k] = (label_mean_area / ground_truth[0].area) * \
                        detections[m, 0].intersection_over_union(ground_truth[0])
                else:
                    iou[m, k] = detections[m, 0].intersection_over_union(ground_truth[0])

        return iou

    compute_similarity_matrix = _rtree_compute_iou_matrix

    def trim_similarity_matrix(self, similarity_matrix, detections, ground_truths, label_mean_area=None):
        r"""Compute an array containing the indices in columns of similarity passing the first trimming.

        Here this is the result of a simple thresholding over IoU.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset,
                if given, it is used to match with *iIoU* instead of *IoU* (c.f. :ref:`iiou`)

        Returns:
            ndarray: An array of dimension (2, N) where each column is a tuple (detection, ground truth) describing
            a potential match. To be more precise, each match-tuple in the array corresponds to a position in the
            similarity matrix which will be used by the match algorithm to compute the final match.

        """
        res = np.stack(np.nonzero(similarity_matrix >= self.threshold))
        return res[:, np.argsort(np.nonzero(similarity_matrix >= self.threshold)[0])]


class MatchEngineEuclideanDistance(MatchEngineBase):
    """Match detection with their ground truth according the their relative distance and the detection confidence score.

    Args:
        threshold (float): The distance threshold at which one considers a potential match as valid
        match_algorithm (str) : Either 'coco', 'xview' or 'non-unitary' to choose the match algorithm

    """

    def __init__(self, threshold, match_algorithm):
        super(MatchEngineEuclideanDistance, self).__init__(match_algorithm)
        self._threshold = 1 - threshold

    @property
    def threshold(self):
        """float: The distance threshold at which one considers a potential match as valid."""
        return 1 - self._threshold

    def compute_similarity_matrix(self, detections, ground_truths, label_mean_area=None):  # noqa: D205,D400
        r"""Compute a partial similarity matrix based on the euclidean distance between all pairs of points with an
        Rtree on detections to speed up computation.

        The difference with :class:`~playground_metrics.match_detections.MatchEnginePointInBox` lies in the
        similarity matrix rough trimming which depends on a threshold rather than on whether a detection (as a point)
        lies within a ground truth polygon (or bounding box).

        The computed matrix is :math:`\mathcal{S} = 1 - \mathcal{D}` with:

        .. math::

            \mathcal{D}_{ij} = \begin{cases} \left\lVert d_i - gt_i \right\rVert_2 &\mbox{if } d_i \in B(gt_i, t)\\
                \inf &\mbox{if }  d_i \notin B(gt_i, t) \end{cases}

        Where :math:`B(gt_i, t)` is a square box centered in :math:`gt_i` of size length :math:`t`.

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray : An similarity matrix (#detections, #ground truth)

        """
        # We prepare the distance matrix (#detection, #gt)
        distance_matrix = np.Inf * np.ones((detections.shape[0], len(ground_truths)))

        detections = self._sort_detection_by_confidence(detections)

        # We construct a Rtree on detections
        def get_bounds(geometry):
            centroid = geometry[0].centroid
            return centroid.x, centroid.y, centroid.x, centroid.y

        rtree_index = _make_rtree(detections, get_bounds)

        for k, ground_truth in enumerate(ground_truths):
            threshold_box = (ground_truth[0].centroid.x - self.threshold, ground_truth[0].centroid.y - self.threshold,
                             ground_truth[0].centroid.x + self.threshold, ground_truth[0].centroid.y + self.threshold)
            overlapping_detections = rtree_index.intersection(threshold_box, objects=False)

            for m in overlapping_detections:
                distance_matrix[m, k] = ground_truth[0].distance(detections[m, 0])
        return 1 - distance_matrix

    def trim_similarity_matrix(self, similarity_matrix, detections, ground_truths, label_mean_area=None):
        r"""Compute an array containing the indices in columns of similarity passing the first trimming.

        Here this is the result of a simple thresholding over the distance matrix.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray: An array of dimension (2, N) where each column is a tuple (detection, ground truth) describing
            a potential match. To be more precise, each match-tuple in the array corresponds to a position in the
            similarity matrix which will be used by the match algorithm to compute the final match.

        """
        res = np.stack(np.nonzero(similarity_matrix >= self._threshold))
        return res[:, np.argsort(np.nonzero(similarity_matrix >= self._threshold)[0])]


class MatchEnginePointInBox(MatchEngineBase):  # noqa: D205,D400
    """Match detection with their ground truth according the their relative distance, whether a detection point is in a
    ground truth box and the detection confidence score.

    Args:
        match_algorithm (str) : Either 'coco', 'xview' or 'non-unitary' to choose the match algorithm

    """

    def __init__(self, match_algorithm):
        super(MatchEnginePointInBox, self).__init__(match_algorithm)

        self._ground_truth_types = (BoundingBox, Polygon)

    def compute_similarity_matrix(self, detections, ground_truths, label_mean_area=None):  # noqa: D205,D400
        r"""Compute a partial similarity matrix based on the euclidean distance between all pairs of points with an
        Rtree on detections to speed up computation.

        The difference with :class:`~playground_metrics.match_detections.MatchEngineEuclideanDistance` lies in the
        similarity matrix rough trimming which depends on whether a detection (as a point) lies within a ground truth
        polygon (or bounding box) rather than on a threshold.

        The computed matrix is :math:`\mathcal{S} = 1 - \mathcal{D}` with:

        .. math::

            \mathcal{D}_{ij} = \left\lVert d_i - gt_i \right\rVert_2

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray : An similarity matrix (#detections, #ground truth)

        """
        # We prepare the distance matrix (#detection, #gt)
        distance_matrix = np.Inf * np.ones((detections.shape[0], len(ground_truths)))

        detections = self._sort_detection_by_confidence(detections)

        # We construct a Rtree on detections
        def get_bounds(geometry):
            centroid = geometry[0].centroid
            return centroid.x, centroid.y, centroid.x, centroid.y

        rtree_index = _make_rtree(detections, get_bounds)

        for k, ground_truth in enumerate(ground_truths):
            overlapping_detections = rtree_index.intersection(ground_truth[0].bounds, objects=False)

            for m in overlapping_detections:
                distance_matrix[m, k] = ground_truth[0].distance(detections[m, 0])
        return 1 - distance_matrix

    def trim_similarity_matrix(self, similarity_matrix, detections, ground_truths, label_mean_area=None):
        r"""Compute an array containing the indices in columns of similarity passing the first trimming.

        Here a detection/ground truth pair is kept if the detection
        :class:`~playground_metrics.utils.geometry_utils.geometry.Point` is within the ground truth
        :class:`~playground_metrics.utils.geometry_utils.geometry.BoundingBox` or
        :class:`~playground_metrics.utils.geometry_utils.geometry.Polygon`

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray: An array of dimension (2, N) where each column is a tuple (detection, ground truth) describing
            a potential match. To be more precise, each match-tuple in the array corresponds to a position in the
            similarity matrix which will be used by the match algorithm to compute the final match.

        """
        potential = np.stack(np.nonzero(similarity_matrix != -np.Inf))[:, np.argsort(np.nonzero(similarity_matrix !=
                                                                                                -np.Inf)[0])]
        trim = []
        for i in range(potential.shape[1]):
            r, c = potential[:, i]
            if detections[r, 0].intersection(ground_truths[c, 0]).is_empty:
                trim.append(i)

        return np.delete(potential, trim, axis=1)


class MatchEngineConstantBox(MatchEngineBase):  # noqa: D205,D400
    """Match detection with their ground truth according the IoU computed on fixed-size
    bounding boxes around detection and ground truth points and the detection confidence score.

    Args:
        threshold (float): The IoU threshold at which one considers a potential match as valid
        match_algorithm (str) : Either 'coco', 'xview' or 'non-unitary' to choose the match algorithm
        bounding_box_size (float): The fixed-size bounding box size

    """

    def __init__(self, threshold, match_algorithm, bounding_box_size):
        super(MatchEngineConstantBox, self).__init__(match_algorithm)
        self.bounding_box_size = bounding_box_size
        self.threshold = threshold

    def compute_similarity_matrix(self, detections, ground_truths, label_mean_area=None):  # noqa: D205,D400
        r"""Compute a parial similarity matrix based on the intersection-over-union between all pairs of constant-sized
        bounding box around points with an Rtree on detections to speed up computation.

        Args:
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray : An IoU matrix (#detections, #ground truth)

        """
        # We prepare the distance matrix (#detection, #gt)
        similarity_matrix = np.zeros((detections.shape[0], len(ground_truths)))

        detections = self._sort_detection_by_confidence(detections)

        # We construct a Rtree on detections
        def get_bounds(geometry):
            return BoundingBox(geometry[0].centroid.x - (self.bounding_box_size // 2),
                               geometry[0].centroid.y - (self.bounding_box_size // 2),
                               geometry[0].centroid.x + (self.bounding_box_size // 2),
                               geometry[0].centroid.y + (self.bounding_box_size // 2)).bounds

        rtree_index = _make_rtree(detections, get_bounds)

        for k, ground_truth in enumerate(ground_truths):
            ground_truth_box = BoundingBox(ground_truth[0].centroid.x - (self.bounding_box_size // 2),
                                           ground_truth[0].centroid.y - (self.bounding_box_size // 2),
                                           ground_truth[0].centroid.x + (self.bounding_box_size // 2),
                                           ground_truth[0].centroid.y + (self.bounding_box_size // 2))
            overlapping_detections = rtree_index.intersection(ground_truth_box.bounds, objects=False)

            for m in overlapping_detections:
                detection_box = BoundingBox(detections[m, 0].centroid.x - (self.bounding_box_size // 2),
                                            detections[m, 0].centroid.y - (self.bounding_box_size // 2),
                                            detections[m, 0].centroid.x + (self.bounding_box_size // 2),
                                            detections[m, 0].centroid.y + (self.bounding_box_size // 2))
                similarity_matrix[m, k] = ground_truth_box.intersection_over_union(detection_box)
        return similarity_matrix

    def trim_similarity_matrix(self, similarity_matrix, detections, ground_truths, label_mean_area=None):
        r"""Compute an array containing the indices in columns of similarity passing the first trimming.

        Here this is the result of a simple thresholding over the intersection-over-union matrix.

        Args:
            similarity_matrix: The similarity matrix between detections and ground truths : dimension (#detection, #gt)
            detections (ndarray, list) : A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[BoundingBox, confidence]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[Polygon, confidence]``
                * Points for a given class where each row is a detection stored as:
                  ``[Point, confidence]``

            ground_truths (ndarray,list) : A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[BoundingBox]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[Polygon]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[Point]``

            label_mean_area (float) : Optional, default to ``None``. The mean area for each label in the dataset.

        Returns:
            ndarray: An array of dimension (2, N) where each column is a tuple (detection, ground truth) describing
            a potential match. To be more precise, each match-tuple in the array corresponds to a position in the
            similarity matrix which will be used by the match algorithm to compute the final match.

        """
        res = np.stack(np.nonzero(similarity_matrix >= self.threshold))
        return res[:, np.argsort(np.nonzero(similarity_matrix >= self.threshold)[0])]



import numpy as np

from playground_metrics.map_metric import MeanAveragePrecisionMetric


class MeanFBetaAtThresholds:
    r"""Class to compute mean F-beta score (averaged and per class) at list of thresholds for detection task.

    Notes:
        Special cases are resolved as following:

        * if no predictions neither ground truth, then f2 = 1 for all classes
        * if no predictions but there is a ground truth, f2 = 0 for all classes
        * if there are predictions, we compute f2 from precision and recall using
          :class:`~playground_metrics.map_metric.MeanAveragePrecisionMetric`

    Args:
        beta (int/float): F-beta scoring parameter
        thresholds (list/tuple of float): List/tuple of similarity thresholds for which we consider a valid
            match between detection and ground truth. For example, `[0.5, 0.7, 0.9]`.
        **kwargs: kwargs to configure internal :class:`~playground_metrics.map_metric.MeanAveragePrecisionMetric`

    Raises:
        ValueError: If ``thresholds`` is not a list/tuple or if ``beta`` is not a positive float

    Attributes:
        score (float) : Mean F-beta score computed by ``compute()`` from accumulated values

    """

    def __init__(self, beta, thresholds, **kwargs):
        """Initialize instance."""
        if not isinstance(thresholds, (tuple, list)):
            raise TypeError("Argument thresholds should be list or tuple, but given {}".format(type(thresholds)))
        if not (isinstance(beta, (int, float)) and beta > 0):
            raise ValueError("Argument beta should be positive float")

        self._map_computers = [
            MeanAveragePrecisionMetric(threshold=t, **kwargs) for t in thresholds
        ]
        self.beta = beta
        self.score = self.score_per_class = self._internal_score_per_class = self._counter = None
        self.reset()

    def reset(self):
        r"""Reset all intermediate and return values to their initial value.

        If ``reset()`` is not called in-between two ``compute()`` call, the values returned by ``compute()`` will take
        into account the entire prediction stack, not just the predictions in-between the two ``compute()`` calls.
        """
        for map_computer in self._map_computers:
            map_computer.reset()
        self.score = 0.0
        self.score_per_class = {}
        self._internal_score_per_class = {}
        self._counter = 0  # counter for seen tiles

    def update(self, detections, ground_truths):
        """Accumulate values necessary to compute mAP with detections and ground truths of a single image.

        Arguments are same as in :meth:`~playground_metrics.map_metric.MeanAveragePrecisionMetric.update`

        """
        detections = np.array(detections, copy=False)
        ground_truths = np.array(ground_truths, copy=False)

        self._counter += 1

        if detections.size == 0:
            if ground_truths.size == 0:
                for label in self._internal_score_per_class:
                    self._internal_score_per_class[label] += 1 * len(self._map_computers)
                # all done -> exit update
                return

        for map_computer in self._map_computers:
            map_computer.reset()
            map_computer.update(detections=detections, ground_truths=ground_truths)
            map_computer.compute()

            # At first we need to discover new classes if any appeared
            for label in map_computer.ground_truth_labels:
                if label not in self._internal_score_per_class:
                    # unseen class -> all previous tiles are correctly predicted for this class
                    self._internal_score_per_class[label] = (self._counter - 1) * len(self._map_computers)

            for label in self._internal_score_per_class:
                if label not in map_computer.ground_truth_labels:
                    f_beta = 1.0
                else:
                    precision = map_computer.precision_per_class[label]
                    recall = map_computer.recall_per_class[label]

                    if not (np.isnan(precision) or np.isnan(recall)):
                        f_beta = self._compute_fbeta(precision, recall)
                    else:
                        f_beta = 0.0
                self._internal_score_per_class[label] += f_beta

    def compute(self):
        """Compute the F-beta score averaged over all classes and thresholds according to the accumulated values.

        Returns:
            float : mean F-beta at thresholds

        """
        if not self._internal_score_per_class:
            return np.nan

        self.score = 0.0
        for label in self._internal_score_per_class:
            self.score_per_class[label] = self._internal_score_per_class[label]
            self.score_per_class[label] /= 1.0 * self._counter  # ~ /= len(tiles)
            self.score_per_class[label] /= 1.0 * len(self._map_computers)  # ~ /= len(thresholds)
            self.score += self.score_per_class[label]

        self.score /= 1.0 * len(self._internal_score_per_class)  # ~ /= n_classes
        return self.score

    def _compute_fbeta(self, precision, recall):
        return (1.0 + self.beta ** 2) * precision * recall / (self.beta ** 2 * precision + recall + 1e-20)



from collections.abc import Sized

import numpy as np
import shapely.ops
import shapely.geometry

from ..exception import InvalidGeometryError
from .geometry import Point, Polygon, BoundingBox, Geometry


# Helpers functions
def get_type_and_convert(input_array, trim_invalid_geometry=False, autocorrect_invalid_geometry=False):
    r"""Automatically find the geometry type from the input array shape and convert it to a geometry array.

    Args:
        input_array (ndarray, list):

            * A ndarray of detections stored as:

                * Bounding boxes for a given class where each row is a detection stored as:
                  ``[x_min, y_min, x_max, y_max, confidence, label]``
                * Polygons for a given class where each row is a detection stored as:
                  ``[[[outer_ring], [inner_rings]], confidence, label]``
                * Points for a given class where each row is a detection stored as:
                  ``[x, y, confidence, label]``

            * A ndarray of ground truth stored as:

                * Bounding boxes for a given class where each row is a ground truth stored as:
                  ``[x_min, y_min, x_max, y_max, label]``
                * Polygons for a given class where each row is a ground truth stored as:
                  ``[[[outer_ring], [inner_rings]], label]``
                * Points for a given class where each row is a ground truth stored as:
                  ``[x, y, label]``

        trim_invalid_geometry (bool): Optional, default to ``False``. If set to ``True`` conversion will ignore invalid
            geometries and leave them out of ``output_array``. This means that the function will return an array where
            ``output_array.shape[0] <= input_array.shape[0]``.  If set to ``False``, an invalid geometry will raise an
            :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`.

        autocorrect_invalid_geometry (Bool): Optional, default to ``False``. Whether to attempt correcting a faulty
            geometry to form a valid one. If set to ``True`` and the autocorrect attempt is unsuccessful, it falls back
            to the behaviour defined in ``trim_invalid_geometry``.

    Note:
        * Polygon auto-correction only corrects self-crossing exterior rings, in which case it creates one Polygon
          out of every simple ring which might be extracted from the original Polygon exterior.
        * Polygon auto-correction will systematically fail on Polygons with at least one inner ring.

    Returns:
        (str, ndarray): A tuple of output containing:

        * The BaseGeometry type as a string which may be either ``"point"``, ``"polygon"`` or ``"bbox"``
        * A geometry ndarray where each row contains a geometry followed by optionally confidence and a label
          e.g.: ``[BaseGeometry, (confidence), label]``

    Raises:
        ValueError: If ``input_array`` have invalid dimensions.

    """
    input_array = np.array(input_array, dtype=np.dtype('O'))
    # If nothing pass here
    if input_array.size == 0:
        return 'undefined', input_array

    if len(input_array.shape) < 2:
        # If we have less than 2 dimensions, pull the plug its useless
        raise ValueError('Invalid array number of dimensions: '
                         'Expected a 2D array, found {}D.'.format(len(input_array.shape)))

    elif len(input_array.shape) == 2:
        # It might just be anything at this point. We need to check the second dimension value to decide
        if input_array.shape[1] < 2:
            # If it's less than 2, pull the plug its useless
            raise ValueError('Invalid array second dimension: '
                             'Expected at least 2, found {}.'.format(input_array.shape[1]))
        elif input_array.shape[1] == 2:
            # It's a Polygon ndarray
            type_ = 'polygon'
            # Convert rings to Polygon
            input_array = convert_to_polygon(input_array, trim_invalid_geometry=trim_invalid_geometry,
                                             autocorrect_invalid_geometry=autocorrect_invalid_geometry)
        elif input_array.shape[1] == 3:
            # It might be either a Polygon ndarray or a Point ndarray
            # We check the first element to decide. One might argue that it is not robust to a mixed-type input array,
            # that is true but the conversion functions implicitly assumes that the input array is of fixed-type for
            # performances issues.
            if isinstance(input_array[0, 0], Sized):
                # Reasonable guess is the first element is a list or a ndarray -> It's a Polygon ndarray
                type_ = 'polygon'
                # Convert rings to Polygon
                input_array = convert_to_polygon(input_array, trim_invalid_geometry=trim_invalid_geometry,
                                                 autocorrect_invalid_geometry=autocorrect_invalid_geometry)
            else:
                # Resonable guess is the first element is a number -> It's a Point ndarray
                type_ = 'point'
                input_array = convert_to_point(input_array, trim_invalid_geometry=trim_invalid_geometry)
        elif input_array.shape[1] == 4:
            # It's a Point ndarray
            type_ = 'point'
            input_array = convert_to_point(input_array, trim_invalid_geometry=trim_invalid_geometry)
        elif 7 > input_array.shape[1] > 4:
            # It's a BoundingBox ndarray
            type_ = 'bbox'
            input_array = convert_to_bounding_box(input_array, trim_invalid_geometry=trim_invalid_geometry)
        else:
            raise ValueError('Invalid array second dimension: '
                             'Expected less than 6, found {}.'.format(input_array.shape[1]))

    elif len(input_array.shape) == 3:
        # It's the weirdest Polygon and tuple class corner case ! This really is not a drill !
        # It's a Polygon ndarray
        type_ = 'polygon'
        # Convert rings to Polygon
        input_array = convert_to_polygon(input_array, trim_invalid_geometry=trim_invalid_geometry,
                                         autocorrect_invalid_geometry=autocorrect_invalid_geometry)

    elif len(input_array.shape) == 5:
        # It's the mildly weird Polygon corner case ! This is not a drill !
        # It's a Polygon ndarray
        type_ = 'polygon'
        # Convert rings to Polygon
        input_array = convert_to_polygon(input_array, trim_invalid_geometry=trim_invalid_geometry,
                                         autocorrect_invalid_geometry=autocorrect_invalid_geometry)

    else:
        # If we have neither 2 dimensions nor 5 (in a weird polygon corner case), pull the plug its useless
        raise ValueError('Invalid array number of dimensions: '
                         'Expected a 2D array, found {}D.'.format(len(input_array.shape)))

    return type_, input_array


def convert_to_polygon(input_array, trim_invalid_geometry=False, autocorrect_invalid_geometry=False):
    r"""Convert an input array to a Polygon array.

    Args:
        input_array (ndarray, list): A ndarray of Polygons optionally followed by a confidence value and/or a label
            where each row is: ``[[[outer_ring], [inner_rings]], (confidence), (label)]``
        trim_invalid_geometry (bool): Optional, default to ``False``. If set to ``True`` conversion will ignore invalid
            geometries and leave them out of ``output_array``. This means that the function will return an array where
            ``output_array.shape[0] <= input_array.shape[0]``.  If set to ``False``, an invalid geometry will raise an
            :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`.
        autocorrect_invalid_geometry (Bool): Optional, default to ``False``. Whether to attempt correcting a faulty
            geometry to form a valid one. If set to ``True`` and the autocorrect attempt is unsuccessful, it falls back
            to the behaviour defined in ``trim_invalid_geometry``.

    Note:
        * Polygon auto-correction only corrects self-crossing exterior rings, in which case it creates one Polygon
          out of every simple ring which might be extracted from the original Polygon exterior.
        * Polygon auto-correction will systematically fail on Polygons with at least one inner ring.

    Returns:
        ndarray: A Polygon ndarray where each row contains a geometry followed by optionally confidence and a label
        e.g.: ``[Polygon, (confidence), (label)]``

    Raises:
        ValueError: If ``input_array`` have invalid dimensions.

    """
    input_array = np.array(input_array, dtype=np.dtype('O'))
    if input_array.size == 0:
        return 'undefined', input_array

    if (len(input_array.shape) == 1 or len(input_array.shape) > 2) and \
            (not len(input_array.shape) == 5 and not len(input_array.shape) == 3):
        raise ValueError('Invalid array number of dimensions: '
                         'Expected a 2D array, found {}D.'.format(len(input_array.shape)))

    if len(input_array.shape) == 5 and not input_array.shape[4] == 2:
        raise ValueError('Invalid array fifth dimension: '
                         'Expected 2, found {}.'.format(len(input_array.shape)))

    elif len(input_array.shape) == 3 and not input_array.shape[2] == 1:
        raise ValueError('Invalid array third dimension: '
                         'Expected 1, found {}.'.format(len(input_array.shape)))

    object_array = np.ndarray((input_array.shape[0], input_array.shape[1]), dtype=np.dtype('O'))
    to_trim = []
    to_add = []
    for i in range(input_array.shape[0]):
        try:
            if len(input_array[i, 0]) > 1:
                list_ = [Polygon(input_array[i, 0][0], *input_array[i, 0][1:])]
                list_.extend(input_array[i, 1:])
            else:
                list_ = [Polygon(input_array[i, 0][0])]
                list_.extend(input_array[i, 1:])
            object_array[i] = np.array(list_, dtype=np.dtype('O'))
        except InvalidGeometryError as e:
            if autocorrect_invalid_geometry:
                if len(shapely.geometry.Polygon(input_array[i, 0][0], input_array[i, 0][1:]).interiors) > 0:
                    raise e
                ext = shapely.geometry.Polygon(input_array[i, 0][0], input_array[i, 0][1:]).exterior
                shapely_polygon_list = \
                    shapely.ops.polygonize(shapely.geometry.LineString(ext.coords[:] + ext.coords[0:1]).intersection(
                        shapely.geometry.LineString(ext.coords[:] + ext.coords[0:1])))
                polygon_list = []
                try:
                    for polygon in shapely_polygon_list:
                        list_ = [Polygon(polygon.exterior.coords)]
                        list_.extend(input_array[i, 1:])
                        polygon_list.append(list_)
                except InvalidGeometryError:
                    if trim_invalid_geometry:
                        to_trim.append(i)
                    else:
                        raise e
                else:
                    to_trim.append(i)
                    to_add.extend(polygon_list)
            else:
                if trim_invalid_geometry:
                    to_trim.append(i)
                else:
                    raise e
    return np.concatenate((np.delete(object_array, to_trim, axis=0),
                           np.array(to_add).reshape(-1, object_array.shape[1])),
                          axis=0)


def convert_to_bounding_box(input_array, trim_invalid_geometry=False, autocorrect_invalid_geometry=False):
    r"""Convert an input array to a BoundingBox array.

    Args:
        input_array (ndarray, list): A ndarray of BoundingBox optionally followed by a confidence value and/or a label
            where each row is: ``[xmin, ymin, xmax, ymax, (confidence), (label)]``
        trim_invalid_geometry (bool): Optional, default to ``False``. If set to ``True`` conversion will ignore invalid
            geometries and leave them out of ``output_array``. This means that the function will return an array where
            ``output_array.shape[0] <= input_array.shape[0]``.  If set to ``False``, an invalid geometry will raise an
            :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`.
        autocorrect_invalid_geometry (Bool): Optional, default to ``False``. Doesn't do anything, introduced to unify
            convert functions interfaces.

    Returns:
        ndarray: A BoundingBox ndarray where each row contains a geometry followed by optionally confidence and a label
        e.g.: ``[BoundingBox, (confidence), (label)]``

    Raises:
        ValueError: If ``input_array`` have invalid dimensions.

    """
    input_array = np.array(input_array, dtype=np.dtype('O'))
    if input_array.size == 0:
        return 'undefined', input_array
    if len(input_array.shape) == 1 or len(input_array.shape) > 2:
        raise ValueError('Invalid array number of dimensions: '
                         'Expected a 2D array, found {}D.'.format(len(input_array.shape)))
    object_array = np.ndarray((input_array.shape[0], input_array.shape[1] - 3), dtype=np.dtype('O'))
    to_trim = []
    for i in range(input_array.shape[0]):
        try:
            list_ = [BoundingBox(*input_array[i, :4])]
            list_.extend(input_array[i, 4:])
            object_array[i] = np.array(list_, dtype=np.dtype('O'))
        except InvalidGeometryError as e:
            if trim_invalid_geometry:
                to_trim.append(i)
            else:
                raise e

    return np.delete(object_array, to_trim, axis=0)


def convert_to_point(input_array, trim_invalid_geometry=False, autocorrect_invalid_geometry=False):
    r"""Convert an input array to a Point array.

    Args:
        input_array (ndarray, list): A ndarray of Point optionally followed by a confidence value and/or a label
            where each row is: ``[x, y, (confidence), (label)]``
        trim_invalid_geometry (bool): Optional, default to ``False``. If set to ``True`` conversion will ignore invalid
            geometries and leave them out of ``output_array``. This means that the function will return an array where
            ``output_array.shape[0] <= input_array.shape[0]``.  If set to ``False``, an invalid geometry will raise an
            :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`.
        autocorrect_invalid_geometry (Bool): Optional, default to ``False``. Doesn't do anything, introduced to unify
            convert functions interfaces.

    Returns:
        ndarray: A Point ndarray where each row contains a geometry followed by optionally confidence and a label
        e.g.: ``[Point, (confidence), (label)]``

    Raises:
        ValueError: If ``input_array`` have invalid dimensions.

    """
    input_array = np.array(input_array, dtype=np.dtype('O'))
    if input_array.size == 0:
        return 'undefined', input_array
    if len(input_array.shape) == 1 or len(input_array.shape) > 2:
        raise ValueError('Invalid array number of dimensions: '
                         'Expected a 2D array, found {}D.'.format(len(input_array.shape)))
    object_array = np.ndarray((input_array.shape[0], input_array.shape[1] - 1), dtype=np.dtype('O'))
    to_trim = []
    for i in range(input_array.shape[0]):
        try:
            list_ = [Point(*input_array[i, :2])]
            list_.extend(input_array[i, 2:])
            object_array[i] = np.array(list_, dtype=np.dtype('O'))
        except InvalidGeometryError as e:
            if trim_invalid_geometry:
                to_trim.append(i)
            else:
                raise e

    return np.delete(object_array, to_trim, axis=0)



import numpy as np
import shapely.geometry

from playground_metrics.utils.geometry_utils.base_geometry import BaseGeometry, AbstractBaseGeometry
from playground_metrics.utils.exception import ShapelySpecificTypeError, InvalidGeometryError


def geometry_factory(geom):
    r"""Convert a shapely.geometry object into either a BoundingBox, a Polygon or a Point object.

    Args:
        geom (:class:`shapely.geometry.BaseGeometry`): The shapely :class:`~shapely.geometry.BaseGeometry` to be
            converted into a :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` object.

    Returns:
        :class:`~map_metric_api.utils.geometry_utils.geometry.Geometry`: A
        :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` object.

    Raises:
        :exc:`~playground_metrics.utils.exception.ShapelySpecificTypeError`: If the provided shapely.geometry object has
            no reliable correspondence with any :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` type
            (e.g. if provided with a :class:`shapely.geometry.MultiPolygon`)

    """
    if not geom.is_empty and isinstance(geom, shapely.geometry.Polygon):  # Might as well be a genuine polygon or a box
        if shapely.geometry.box(*geom.bounds).equals(geom):  # It's a box
            return BoundingBox(*geom.bounds)
        else:  # It is a genuine polygon, we don't take any risk anyway
            return Polygon(list(iter(geom.exterior.coords)), *[list(iter(interior.coords))
                                                               for interior in geom.interiors])
    elif not geom.is_empty and isinstance(geom, shapely.geometry.Point):  # It's a simple point
        return Point(geom.x, geom.y)
    else:
        raise ShapelySpecificTypeError('Resulting shapely.geometry ({}) '
                                       'type may not reliably be mapped to a '
                                       'valid playground_metrics BaseGeometry'.format(geom))


class Geometry(BaseGeometry):
    r"""A concrete realisation of :class:`~playground_metrics.utils.geometry_utils.base_geometry.BaseGeometry`.

    It implements the backward mapping in :meth:`from_shapely` and acts a base Geometry type for undefined Geometries
    (such as an empty Geometry).
    """

    __slots__ = ()

    @classmethod
    def from_shapely(cls, geom):
        r"""Perform the inverse mapping from a shapely object to a Geometry type.

        To be more precise, :class:`~playground_metrics.utils.geometry_utils.base_geometry.BaseGeometry` subclasses
        implement a forward mapping from a Geometry type to a shapely object. However, because geometric operations are
        not guaranteed to keep the Geometric type (e.g. intersection between a Polygon and a Point is a Point, not a
        Polygon), and because all the actual operations are performed by shapely, it is necessary to be able to perform
        the backward mapping to transform an operation result back into valid Geometry type.

        It internally calls upon the mapping implementation in :func:`geometry_factory` and catches
        :exc:`~playground_metrics.utils.exception.ShapelySpecificTypeError` to instantiate
        a :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` in case of a undefined type.

        Args:
            geom (:class:`shapely.geometry.BaseGeometry`): The shapely :class:`~shapely.geometry.BaseGeometry` to be
                converted into a :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` object.

        Returns:
            :class:`~map_metric_api.utils.geometry_utils.base_geometry.BaseGeometry`: The backward mapping resulting
            Geometry type object.

        """
        try:
            return geometry_factory(geom)
        except ShapelySpecificTypeError:
            return Geometry(geom)


class Point(Geometry):
    r"""A point geometry.

    Args:
        x (float): The x coordinate
        y (float): The y coordinate

    Attributes:
        x (float): The x coordinate
        y (float): The y coordinate

    """

    __slots__ = ()

    def __init__(self, x, y):
        shapely_internal = shapely.geometry.Point(x, y)
        super(Point, self).__init__(shapely_internal, x=x, y=y)

    def distance(self, *others):  # noqa: D205,D400
        r"""Return the euclidean distance between the object centroid and the `other` object centroid or
        a tuple of distance if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute centroid distance with

        Returns:
            (float, tuple): The centroid distance or a tuple of centroid distance depending on the number of other
            geometries

        """
        if len(others) == 1:
            return self._internal.distance(others[0].centroid)
        else:
            return tuple(self._internal.distance(other.centroid) for other in others)


class Polygon(Geometry):
    r"""A polygon geometry.

    Args:
        outer (list, ndarray): An array of (x, y) couples containing the outer ring of the polygon (i.e. the shell)
        *inners (tuple): Optional. One or more array similar to `outer` containing the inner rings of polygon
            (i.e. the holes)

    Attributes:
        shell (ndarray): An array of (x, y) couples containing the outer ring of the polygon (i.e. the shell)
        holes (ndarray): One or more array similar to `outer` containing the inner rings of polygon (i.e. the holes)
            or an empty array.

    """

    __slots__ = ()

    def __init__(self, outer, *inners):
        outer = np.array(outer)
        inners = np.array(inners)
        shapely_internal = shapely.geometry.Polygon(outer, inners)
        super(Polygon, self).__init__(shapely_internal, shell=outer, holes=inners)


class FastBoundingBox(AbstractBaseGeometry):
    r"""A fast bounding box geometry.

    It differs from the canonical :class:`~playground_metrics.utils.geometry_utils.geometry.ShapelyBoundingBox` by
    delaying the shapely internal creation only to when it really is needed and tries to perform fast box-to-box
    operation whenever possible.

    Args:
        xmin (float): The left-most coordinate
        ymin (float): The upper-most coordinate
        xmax (float): The right-most coordinate
        ymax (float): The lower-most coordinate

    Raises:
        :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`: If ``xmax <= xmin`` or ``ymax <= ymin``.

    Attributes:
        xmin (float): The left-most coordinate
        ymin (float): The upper-most coordinate
        xmax (float): The right-most coordinate
        ymax (float): The lower-most coordinate
        area (float): The area of the object.

    """

    __slots__ = '__internal', 'xmin', 'xmax', 'ymin', 'ymax', 'area'

    def __init__(self, xmin, ymin, xmax, ymax):
        if (xmax <= xmin or ymax <= ymin) and (not xmin == xmax == ymin == ymax == np.inf):
            raise InvalidGeometryError('Invalid box coordinates (xmax ({xmax}) <= xmin ({xmin}) or '
                                       'ymax ({ymax}) <= ymin ({ymin}))'.format(xmin=xmin, ymin=ymin,
                                                                                xmax=xmax, ymax=ymax))
        # Delay _internal creation for as long as it is possible
        self.__internal = None
        super(FastBoundingBox, self).__init__(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)

        # Fast coordinates access
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

        # Fast area pre-computation
        self.area = (self.xmax - self.xmin) * (self.ymax - self.ymin)

    @property
    def width(self):
        """float: The box's width."""
        return self.xmax - self.xmin

    @property
    def height(self):
        """float: The box's height."""
        return self.ymax - self.ymin

    @property
    def _internal(self):
        # When we do need _internal, if it was not created, we have to now.
        if self.__internal is None:
            self.__internal = shapely.geometry.box(self.xmin, self.ymin, self.xmax, self.ymax)
        return self.__internal

    @property
    def is_empty(self):
        r"""bool: Whether the feature's `interior` and `boundary` (in point set terms) coincide with the empty set."""
        return (self.xmin == self.xmax) or (self.ymin == self.ymax)

    @property
    def bounds(self):
        r"""(float, float, float, float): A ``(minx, miny, maxx, maxy)`` tuple of coordinates that bound the object."""
        return self.xmin, self.ymin, self.xmax, self.ymax

    @property
    def centroid(self):  # noqa: D205,D400
        r""":class:`~playground_metrics.utils.geometry_utils.geometry.Point`:
        The objects centroid as a Geometry type object.
        """
        return Point(0.5 * (self.xmin + self.xmax), 0.5 * (self.ymin + self.ymax))

    def equals(self, other):  # noqa: D205,D400
        r"""Return ``True`` if the set-theoretic `boundary`, `interior`, and `exterior` of the object coincide with
        those of the other.

        Args:
            other (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometry to test equality with

        Returns:
            bool: Whether the subsets of the 2D plane described by the 2 geometries are equal.

        """
        if hasattr(other, 'xmin') and hasattr(other, 'ymin') and hasattr(other, 'xmax') and hasattr(other, 'ymax'):
            # Try fast box-to-box equals
            return self.xmin == other.xmin and self.ymin == other.ymin and \
                self.xmax == other.xmax and self.ymax == other.ymax
        else:  # If not possible, fallback to shapely equals
            return self._internal.equals(other)

    def distance(self, *others):  # noqa: D205,D400
        r"""Return the euclidean distance between the object centroid and the `other` object centroid or
        a tuple of distance if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute centroid distance with

        Returns:
            (float, tuple): The centroid distance or a tuple of centroid distance depending on the number of other
            geometries

        """
        if len(others) == 1:
            return self.centroid.distance(others[0].centroid)
        else:
            return tuple(self.centroid.distance(other.centroid) for other in others)

    def _intersection_bounding_box(self, other):
        """Compute the intersection of self with bounding boxes.

        Args:
            self (FastBoundingBox, ShapelyBoundingBox): A bounding box type geometry
            other (FastBoundingBox, ShapelyBoundingBox): A bounding box type geometry

        Returns:
            (float, float, float, float) : The intersection coordinates

        """
        # Determine the (x, y)-coordinates of the intersection rectangle
        x_min = max(self.xmin, other.xmin)
        y_min = max(self.ymin, other.ymin)
        x_max = min(self.xmax, other.xmax)
        y_max = min(self.ymax, other.ymax)

        return x_min, y_min, max(x_max, x_min), max(y_max, y_min)

    def _intersection(self, other):
        if hasattr(other, 'xmin') and hasattr(other, 'ymin') and hasattr(other, 'xmax') and hasattr(other, 'ymax'):
            # Try fast box-to-box intersection
            try:
                return FastBoundingBox(*self._intersection_bounding_box(other))
            except InvalidGeometryError:
                return shapely.geometry.MultiPolygon()
        else:  # If not possible, fallback to shapely intersection
            return self._internal.intersection(other)

    def intersection(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection of this object with the `other` geometric object or a tuple
        of intersection if multiple input were given.

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection with

        Returns:
            (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`, tuple): The intersection
            or a tuple of intersection depending on the number of other geometries

        """
        if len(others) == 1:
            return self.from_shapely(self._intersection(others[0]))
        else:
            return tuple(self.from_shapely(self._intersection(other)) for other in others)

    def intersection_over_union(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection-over-union of this object with the `other` geometric object or
        a tuple of intersection-over-union if multiple input were given (c.f. :ref:`iou`).

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection over union with

        Returns:
            (float, tuple): The intersection over union or a tuple of intersection over union depending on the number
            of other geometries

        """
        if self.xmin == self.xmax == self.ymin == self.ymax == np.inf:
            if len(others) == 1:
                return 0.0
            else:
                return tuple(0.0 for _ in others)

        out = [None] * len(others)

        if len(others) == 1:
            inters = (self._intersection(others[0]), )
        else:
            inters = tuple(self._intersection(other) for other in others)

        for i, (other, inter) in enumerate(zip(others, inters)):

            if inter.is_empty or inter.area == 0:
                out[i] = 0.0
                continue

            # Compute the intersection over union by taking the intersection area and dividing it by the sum of
            # prediction + ground-truth areas - the intersection area
            iou = inter.area / float(self.area + other.area - inter.area)

            # Return the intersection over union value
            out[i] = iou

        if len(out) == 1:
            return out[0]
        else:
            return tuple(out)

    @classmethod
    def from_shapely(cls, geom):
        r"""Perform the inverse mapping from a shapely object to a Geometry type.

        To be more precise, :class:`~playground_metrics.utils.geometry_utils.base_geometry.BaseGeometry` subclasses
        implement a forward mapping from a Geometry type to a shapely object. However, because geometric operations are
        not guaranteed to keep the Geometric type (e.g. intersection between a Polygon and a Point is a Point, not a
        Polygon), and because all the actual operations are performed by shapely, it is necessary to be able to perform
        the backward mapping to transform an operation result back into valid Geometry type.

        It internally calls upon the mapping implementation in :func:`geometry_factory` and catches
        :exc:`~playground_metrics.utils.exception.ShapelySpecificTypeError` to instantiate
        a :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` in case of a undefined type.

        Args:
            geom (:class:`shapely.geometry.BaseGeometry`): The shapely :class:`~shapely.geometry.BaseGeometry` to be
                converted into a :class:`~playground_metrics.utils.geometry_utils.geometry.Geometry` object.

        Returns:
            :class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`: The backward mapping
            resulting Geometry type object.

        """
        if isinstance(geom, FastBoundingBox):
            return geom
        try:
            return geometry_factory(geom)
        except ShapelySpecificTypeError:
            return Geometry(geom)


class ShapelyBoundingBox(Geometry):
    r"""A bounding box geometry.

    Args:
        xmin (float): The left-most coordinate
        ymin (float): The upper-most coordinate
        xmax (float): The right-most coordinate
        ymax (float): The lower-most coordinate

    Raises:
        :exc:`~playground_metrics.utils.geometry_utils.InvalidGeometryError`: If ``xmax <= xmin`` or ``ymax <= ymin``.

    Attributes:
        xmin (float): The left-most coordinate
        ymin (float): The upper-most coordinate
        xmax (float): The right-most coordinate
        ymax (float): The lower-most coordinate

    """

    def __init__(self, xmin, ymin, xmax, ymax):
        if (xmax <= xmin or ymax <= ymin) and (not xmin == xmax == ymin == ymax == np.inf):
            raise InvalidGeometryError('Invalid box coordinates (xmax ({xmax}) <= xmin ({xmin}) or '
                                       'ymax ({ymax}) <= ymin ({ymin}))'.format(xmin=xmin, ymin=ymin,
                                                                                xmax=xmax, ymax=ymax))
        shapely_internal = shapely.geometry.box(xmin, ymin, xmax, ymax)
        super(ShapelyBoundingBox, self).__init__(shapely_internal, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)

    @property
    def width(self):
        """float: The box's width."""
        return self.xmax - self.xmin

    @property
    def height(self):
        """float: The box's height."""
        return self.ymax - self.ymin

    def intersection_over_union(self, *others):  # noqa: D205,D400
        r"""Return a representation of the intersection-over-union of this object with the `other` geometric object or
        a tuple of intersection-over-union if multiple input were given (c.f. :ref:`iou`).

        Args:
            *others (:class:`~map_metric_api.utils.geometry_utils.base_geometry.AbstractBaseGeometry`): Other
                geometries to compute intersection over union with

        Returns:
            (float, tuple): The intersection over union or a tuple of intersection over union depending on the number
            of other geometries

        """
        if self.xmin == self.xmax == self.ymin == self.ymax == np.inf:
            if len(others) == 1:
                return 0.0
            else:
                return tuple(0.0 for o in others)
        return super(ShapelyBoundingBox, self).intersection_over_union(*others)


BoundingBox = FastBoundingBox



import warnings
from functools import wraps


def deprecated(reason):
    """Deprecate a function by adding a ``DeprecationWarning`` and a **Warning** block in the docstring."""
    def deprecated_decorator(func):
        @wraps(func)
        def deprecated_function(*args, **kwargs):
            warnings.warn('The function {0} is deprecated and may not work or '
                          'disappear in the future'.format(func.__name__), DeprecationWarning)
            return func(*args, **kwargs)

        former_docstring = deprecated_function.__doc__ or ''

        deprecated_function.__doc__ = \
            """
            .. warning::
                The function ``{0}`` is deprecated and may not work anymore or disappear in the future.

                Reason for deprecation: *{1}*

            """.format(func.__name__, reason) + former_docstring
        return deprecated_function
    return deprecated_decorator



import io
import os
import re
from setuptools import setup, find_packages

REQUIRED_PACKAGES = [
    'numpy>=1.15.4',
    'shapely~=1.7',
    'rtree>=0.8.3',
]


base_dir = os.path.dirname(__file__)

with io.open(os.path.join(base_dir, 'playground_metrics/__init__.py'), 'rt', encoding='utf8') as f:
    version = re.search(
        r'__version__ = \'(.*?)\'', f.read(), re.M).group(1)


with open(os.path.join(base_dir, "README.rst")) as f:
    long_description = f.read()


with open(os.path.join(base_dir, "CHANGELOG.rst")) as f:
    changelog = f.read()
    long_description = "\n".join([long_description, changelog])


setup(
    name='playground-metrics',
    version=str(version),
    packages=find_packages(exclude=['tests', 'tests.*', 'docs', 'docs.*']),
    description='Playground framework agnostic detection mAP scoring library.',
    url="https://github.com/airbusgeo/playground-metrics",
    include_package_data=True,
    author='Airbus DS GEO',
    author_email='jeffaudi.airbus@gmail.com',
    license='MIT',
    install_requires=REQUIRED_PACKAGES,
    extras_require={'lint': ['nox',
                             'flake8',
                             'flake8-docstrings',
                             'pep8-naming',
                             'flake8-mutable',
                             'flake8-eradicate',
                             'flake8-comprehensions',
                             'flake8-import-order'],
                    'tests': ['nox',
                              'pytest',
                              'pytest-cov'],
                    'tests_visualisation_utility': ['descartes'],
                    'docs': ['sphinx',
                             'sphinx-rtd-theme']},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    zip_safe=False)
